package solution;

import java.util.Iterator;
import java.util.Stack;
import starter.OrderedSet;

/**
 * A Binary Search Tree implementation of OrderedSet ADT.
 *
 * @param <T> Element type.
 */
public class BstOrderedSet<T extends Comparable<T>> implements OrderedSet<T> {

  private Node<T> root;
  private int numElements;

  /**
   * Construct an empty BstOrderedSet.
   */
  public BstOrderedSet() {
    root = null;
    numElements = 0;
  }

  /**
   * Driver program to visualize/test this implementation.
   *
   * @param args command-line arguments.
   */
  public static void main(String[] args) {
    BstOrderedSet<Integer> bst = new BstOrderedSet<>();
    bst.insert(7);
    bst.insert(2);
    bst.insert(13);
    bst.insert(4);
    bst.insert(10);
    bst.insert(15);
    bst.insert(5);
    bst.insert(8);
    bst.insert(11);
    bst.insert(14);
    bst.insert(17);

    bst.remove(5);
    bst.remove(2);
    bst.remove(13);
  }

  /* recursive find */
  private Node<T> find(Node<T> node, T data) {
    if (node == null) {
      return null;
    }
    if (node.data.compareTo(data) > 0) {
      return find(node.left, data);
    } else if (node.data.compareTo(data) < 0) {
      return find(node.right, data);
    } else {
      return node;
    }
  }

  @Override
  public void insert(T t) {
    // Uses a recursive (private) helper insert
    root = insert(root, t);
  }

  /* inserts given value into subtree rooted at given node
      and returns changed subtree with new node added. */
  private Node<T> insert(Node<T> node, T data) {
    if (node == null) {
      /* If the subtree is empty, return a new node */
      numElements++;
      return new Node<>(data);
    }

    /* Otherwise, recur down the tree */
    if (node.data.compareTo(data) > 0) {
      node.left = insert(node.left, data);
    } else if (node.data.compareTo(data) < 0) {
      node.right = insert(node.right, data);
    }

    /* return the (unchanged) node pointer */
    return node;
  }

  @Override
  public void remove(T t) {
    // Uses a recursive (private) helper insert
    root = remove(root, t);
  }

  // always call as follows: subtree = remove(subtree, data);
  private Node<T> remove(Node<T> node, T t) {
    // if node == null, element isn't present --> don't do anything
    if (node != null) {
      int comparisonResult = t.compareTo(node.data);
      if (comparisonResult < 0) {
        node.left = remove(node.left, t);
      } else if (comparisonResult > 0) {
        node.right = remove(node.right, t);
      } else { // comparisonResult == 0 --> found the element!
        node = removeSubtreeRoot(node);
      }
    }
    return node;
  }

  // always call as follows: subtree = removeSubtreeRoot(subtree)
  // pre: node != null
  private Node<T> removeSubtreeRoot(Node<T> node) {
    if (node.left == null) {
      node = node.right;
      numElements--;
    } else if (node.right == null) {
      node = node.left;
      numElements--;
    } else { // two children:
      T smallestRight = getSmallestElement(node.right);
      node.data = smallestRight;
      node.right = remove(node.right, smallestRight); // takes care of numElements--
    }
    return node;
  }

  // pre: node != null
  private T getSmallestElement(Node<T> node) {
    while (node.left != null) {
      node = node.left;
    }
    return node.data;
  }

  @Override
  public boolean has(T t) {
    return find(root, t) != null;
  }

  @Override
  public int size() {
    return numElements;
  }

  @Override
  public Iterator<T> iterator() {
    return new InorderIterator();
  }

  private static class Node<E> {
    E data;
    Node<E> left;
    Node<E> right;

    Node(E data) {
      this.data = data;
    }
  }

  private class InorderIterator implements Iterator<T> {
    private final Stack<Node<T>> stack;

    InorderIterator() {
      stack = new Stack<>();
      pushLeft(root);
    }

    private void pushLeft(Node<T> curr) {
      while (curr != null) {
        stack.push(curr);
        curr = curr.left;
      }
    }

    @Override
    public boolean hasNext() {
      return !stack.isEmpty();
    }

    @Override
    public T next() {
      Node<T> top = stack.pop();
      pushLeft(top.right);
      return top.data;
    }
  }
}
